/********************************************************************************
** Form generated from reading UI file 'qgs3drendererrulepropswidget.ui'
**
** Created by: Qt User Interface Compiler version 6.11.0
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_QGS3DRENDERERRULEPROPSWIDGET_H
#define UI_QGS3DRENDERERRULEPROPSWIDGET_H

#include <QtCore/QVariant>
#include <QtGui/QIcon>
#include <QtWidgets/QApplication>
#include <QtWidgets/QFormLayout>
#include <QtWidgets/QFrame>
#include <QtWidgets/QGroupBox>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QLabel>
#include <QtWidgets/QLineEdit>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QRadioButton>
#include <QtWidgets/QVBoxLayout>
#include <QtWidgets/QWidget>
#include "qgsscrollarea.h"

QT_BEGIN_NAMESPACE

class Ui_Qgs3DRendererRulePropsWidget
{
public:
    QVBoxLayout *verticalLayout;
    QgsScrollArea *scrollArea;
    QWidget *scrollAreaWidgetContents;
    QVBoxLayout *verticalLayout_2;
    QFormLayout *formLayout;
    QLabel *label_1;
    QLineEdit *editDescription;
    QRadioButton *mFilterRadio;
    QRadioButton *mElseRadio;
    QLabel *label;
    QFrame *filterFrame;
    QHBoxLayout *horizontalLayout;
    QLineEdit *editFilter;
    QPushButton *btnExpressionBuilder;
    QPushButton *btnTestFilter;
    QGroupBox *groupSymbol;

    void setupUi(QWidget *Qgs3DRendererRulePropsWidget)
    {
        if (Qgs3DRendererRulePropsWidget->objectName().isEmpty())
            Qgs3DRendererRulePropsWidget->setObjectName("Qgs3DRendererRulePropsWidget");
        Qgs3DRendererRulePropsWidget->resize(579, 336);
        verticalLayout = new QVBoxLayout(Qgs3DRendererRulePropsWidget);
        verticalLayout->setObjectName("verticalLayout");
        verticalLayout->setContentsMargins(0, 0, 0, 0);
        scrollArea = new QgsScrollArea(Qgs3DRendererRulePropsWidget);
        scrollArea->setObjectName("scrollArea");
        scrollArea->setFrameShape(QFrame::NoFrame);
        scrollArea->setWidgetResizable(true);
        scrollAreaWidgetContents = new QWidget();
        scrollAreaWidgetContents->setObjectName("scrollAreaWidgetContents");
        scrollAreaWidgetContents->setGeometry(QRect(0, 0, 579, 336));
        verticalLayout_2 = new QVBoxLayout(scrollAreaWidgetContents);
        verticalLayout_2->setObjectName("verticalLayout_2");
        verticalLayout_2->setContentsMargins(0, 0, 0, 0);
        formLayout = new QFormLayout();
        formLayout->setObjectName("formLayout");
        formLayout->setFieldGrowthPolicy(QFormLayout::ExpandingFieldsGrow);
        label_1 = new QLabel(scrollAreaWidgetContents);
        label_1->setObjectName("label_1");

        formLayout->setWidget(0, QFormLayout::ItemRole::LabelRole, label_1);

        editDescription = new QLineEdit(scrollAreaWidgetContents);
        editDescription->setObjectName("editDescription");

        formLayout->setWidget(0, QFormLayout::ItemRole::FieldRole, editDescription);

        mFilterRadio = new QRadioButton(scrollAreaWidgetContents);
        mFilterRadio->setObjectName("mFilterRadio");

        formLayout->setWidget(1, QFormLayout::ItemRole::LabelRole, mFilterRadio);

        mElseRadio = new QRadioButton(scrollAreaWidgetContents);
        mElseRadio->setObjectName("mElseRadio");

        formLayout->setWidget(2, QFormLayout::ItemRole::LabelRole, mElseRadio);

        label = new QLabel(scrollAreaWidgetContents);
        label->setObjectName("label");

        formLayout->setWidget(2, QFormLayout::ItemRole::FieldRole, label);

        filterFrame = new QFrame(scrollAreaWidgetContents);
        filterFrame->setObjectName("filterFrame");
        horizontalLayout = new QHBoxLayout(filterFrame);
        horizontalLayout->setObjectName("horizontalLayout");
        horizontalLayout->setContentsMargins(0, 0, 0, 0);
        editFilter = new QLineEdit(filterFrame);
        editFilter->setObjectName("editFilter");

        horizontalLayout->addWidget(editFilter);

        btnExpressionBuilder = new QPushButton(filterFrame);
        btnExpressionBuilder->setObjectName("btnExpressionBuilder");
        QSizePolicy sizePolicy(QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Fixed);
        sizePolicy.setHorizontalStretch(0);
        sizePolicy.setVerticalStretch(0);
        sizePolicy.setHeightForWidth(btnExpressionBuilder->sizePolicy().hasHeightForWidth());
        btnExpressionBuilder->setSizePolicy(sizePolicy);
        QIcon icon;
        icon.addFile(QString::fromUtf8(":/images/themes/default/mIconExpression.svg"), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        btnExpressionBuilder->setIcon(icon);

        horizontalLayout->addWidget(btnExpressionBuilder);

        btnTestFilter = new QPushButton(filterFrame);
        btnTestFilter->setObjectName("btnTestFilter");

        horizontalLayout->addWidget(btnTestFilter);


        formLayout->setWidget(1, QFormLayout::ItemRole::FieldRole, filterFrame);


        verticalLayout_2->addLayout(formLayout);

        groupSymbol = new QGroupBox(scrollAreaWidgetContents);
        groupSymbol->setObjectName("groupSymbol");
        QSizePolicy sizePolicy1(QSizePolicy::Policy::Preferred, QSizePolicy::Policy::Expanding);
        sizePolicy1.setHorizontalStretch(0);
        sizePolicy1.setVerticalStretch(0);
        sizePolicy1.setHeightForWidth(groupSymbol->sizePolicy().hasHeightForWidth());
        groupSymbol->setSizePolicy(sizePolicy1);
        groupSymbol->setCheckable(true);

        verticalLayout_2->addWidget(groupSymbol);

        scrollArea->setWidget(scrollAreaWidgetContents);

        verticalLayout->addWidget(scrollArea);

        QWidget::setTabOrder(scrollArea, editDescription);
        QWidget::setTabOrder(editDescription, mFilterRadio);
        QWidget::setTabOrder(mFilterRadio, editFilter);
        QWidget::setTabOrder(editFilter, btnExpressionBuilder);
        QWidget::setTabOrder(btnExpressionBuilder, btnTestFilter);
        QWidget::setTabOrder(btnTestFilter, mElseRadio);
        QWidget::setTabOrder(mElseRadio, groupSymbol);

        retranslateUi(Qgs3DRendererRulePropsWidget);

        QMetaObject::connectSlotsByName(Qgs3DRendererRulePropsWidget);
    } // setupUi

    void retranslateUi(QWidget *Qgs3DRendererRulePropsWidget)
    {
        Qgs3DRendererRulePropsWidget->setWindowTitle(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Rule Properties", nullptr));
        label_1->setText(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Description", nullptr));
        mFilterRadio->setText(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Filter", nullptr));
        mElseRadio->setText(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Else", nullptr));
        label->setText(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Catch-all for other features", nullptr));
        btnTestFilter->setText(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Test", nullptr));
        groupSymbol->setTitle(QCoreApplication::translate("Qgs3DRendererRulePropsWidget", "Symbol", nullptr));
    } // retranslateUi

};

namespace Ui {
    class Qgs3DRendererRulePropsWidget: public Ui_Qgs3DRendererRulePropsWidget {};
} // namespace Ui

QT_END_NAMESPACE

#endif // UI_QGS3DRENDERERRULEPROPSWIDGET_H
